--[[
	Kui Nameplates
	Kesava-Auchindoun
]]

local addon, ns = ...
local kui = LibStub('Kui-1.0')
local LSM = LibStub('LibSharedMedia-3.0')
local kn = LibStub('AceAddon-3.0'):GetAddon('KuiNameplates')
local slowUpdateTime, critUpdateTime = 1, .1

--KuiNameplatesDebug=true
-- uncommenting this line will display a white box where the default UI's
-- nameplate is rendered, a black background where the pixel-perfect plate from
-- this addon is rendered (if "Fix aliasing" is enabled), the GUID/name
-- status of any visible plates and the detected reaction type of the plate.

kn.font = ''
kn.fontSizes = {}
kn.texSizes  = {}
kn.sizes     = {}
kn.frameList = {}

-- Custom reaction colours
kn.r = {
    { .7, .2, .1 }, -- hated
    { 1, .8, 0 },   -- neutral
    { .2, .6, .1 }, -- friendly
	{ .5, .5, .5 }, -- tapped
}

-- sizes of frame elements
kn.defaultSizes = {
	frame = {
		width    = 110,
		height   = 11,
		twidth   = 55,
		theight  = 7,
		bgOffset = 4  -- inset offset for the frame glow (frame.bg)
	},
	font = {
		large     = 10,
		spellname = 9,
		name      = 9,
		small     = 8
	},
	tex = {
		raidicon = 20,
		targetGlowW = 106,
		targetGlowH = 5,
	}
}

local loadedGUIDs, loadedNames = {}, {}
local targetExists
local profile, uiscale

--------------------------------------------------------------------- globals --
local select, strfind, strsplit, pairs, ipairs, unpack, tinsert, type, floor
    = select, strfind, strsplit, pairs, ipairs, unpack, tinsert, type, floor

------------------------------------------------------------ helper functions --

local function SetFontSize(fs, size)
	if profile.fonts.options.onesize then
		size = kn.fontSizes['name']
	end

	if type(size) == 'string' and fs.size and kn.fontSizes[size] then
		fs.size = size
		size = kn.fontSizes[size]
	end

	local font, _, flags = fs:GetFont()
	fs:SetFont(font, size, flags)
end

local function CreateFontString(self, parent, obj)
	-- store size as a key of kn.fontSizes so that it can be recalled & scaled
	-- correctly. Used by SetFontSize.
	local sizeKey

	obj = obj or {}	
	obj.mono = profile.fonts.options.monochrome
	obj.outline = profile.fonts.options.outline
	obj.size = obj.size or 'name'

	if type(obj.size) == 'string' then
		sizeKey = obj.size
		obj.size = kn.fontSizes[sizeKey]
	end

	if profile.fonts.options.onesize then
		obj.size = kn.fontSizes['name']
	end

	if not obj.font then
		obj.font = kn.font
	end

	local fs = kui.CreateFontString(parent, obj)
	fs.size = sizeKey

	--fs.SetFontSize = SetFontSize
	--fs.SetFontStyle = SetFontStyle

	tinsert(self.fontObjects, fs)
	return fs
end

kn.SetFontSize = SetFontSize
kn.CreateFontString = CreateFontString

------------------------------------------------------------- Frame functions --
-- set colour of health bar according to reaction/threat
local function SetHealthColour(self)
	if self.hasThreat then
		self.health.reset = true
		self.health:SetStatusBarColor(unpack(profile.tank.barcolour))
		return
	end

	local r, g, b = self.oldHealth:GetStatusBarColor()
	if self.health.reset  or
	   r ~= self.health.r or
	   g ~= self.health.g or
	   b ~= self.health.b
	then
		-- store the default colour
		self.health.r, self.health.g, self.health.b = r, g, b
		self.health.reset, self.friend, self.player = nil, nil, nil

		if g > .9 and r == 0 and b == 0 then
			-- friendly NPC
			self.friend = true
			r, g, b = unpack(kn.r[3])
		elseif b > .9 and r == 0 and g == 0 then
			-- friendly player
			self.friend = true
			self.player = true
			r, g, b = 0, .3, .6
		elseif r > .9 and g == 0 and b == 0 then
			-- enemy NPC
			r, g, b = unpack(kn.r[1])
		elseif (r + g) > 1.8 and b == 0 then
			-- neutral NPC
			r, g, b = unpack(kn.r[2])
		elseif b > .9 and (r+g) > 1.06 and (r+g) < 1.1 then
			-- tapped NPC
			r, g, b = unpack(kn.r[4])
		else
			-- enemy player, use default UI colour
			self.player = true
		end

		self.health:SetStatusBarColor(r, g, b)
	end
end

local function SetNameColour(self)
	if self.friend then
		self.name:SetTextColor(unpack(profile.text.friendlyname))
	else
		self.name:SetTextColor(unpack(profile.text.enemyname))
	end
end

local function SetGlowColour(self, r, g, b, a)
	if not r then
		-- set default colour
		r, g, b = 0, 0, 0
	end

	if not a then
		a = .85
	end

    self.bg:SetVertexColor(r, g, b, a)
end

local function StoreFrameGUID(self, guid)
	if not guid then return end
	if self.guid and loadedGUIDs[self.guid] then
		if self.guid ~= guid then
			-- the currently stored guid is incorrect
			loadedGUIDs[self.guid] = nil
		else
			return
		end
	end

	self.guid = guid
	loadedGUIDs[guid] = self

	if loadedNames[self.name.text] == self then
		loadedNames[self.name.text] = nil
	end
end

---------------------------------------------------- Update health bar & text --
local function OnHealthValueChanged(oldBar, curr)
	local frame	= oldBar:GetParent():GetParent().kui
	local min, max	= oldBar:GetMinMaxValues()
	local deficit,    big, sml, condition, display, pattern, rules
	    = max - curr, '',  ''

	frame.health:SetMinMaxValues(min, max)
	frame.health:SetValue(curr)
	
	-- select correct health display pattern
	if frame.friend then
		pattern = profile.hp.friendly
	else
		pattern = profile.hp.hostile
	end

	-- parse pattern into big/sml
	rules = { strsplit(';', pattern) }

	for k, rule in ipairs(rules) do
		condition, display = strsplit(':', rule)

		if condition == '<' then
			condition = curr < max
		elseif condition == '=' then
			condition = curr == max
		elseif condition == '<=' or condition == '=<' then
			condition = curr <= max
		else
			condition = nil
		end

		if condition then
			if display == 'd' then
				big = '-'..kui.num(deficit)
				sml = kui.num(curr)
			elseif display == 'm' then
				big = kui.num(max)
			elseif display == 'c' then
				big = kui.num(curr)
				sml = curr ~= max and kui.num(max)
			elseif display == 'p' then
				big = floor(curr / max * 100)
				sml = kui.num(curr)
			end

			break
		end
	end

	frame.health.p:SetText(big)

	if frame.health.mo then
		frame.health.mo:SetText(sml)
	end
end

------------------------------------------------------- Frame script handlers --
local function OnFrameShow(self)
	local f = self.kui

	-- reset name
	f.name.text = f.oldName:GetText()
	f.name:SetText(f.name.text)

	if profile.hp.mouseover then
		-- force un-highlight
		f.highlighted = true
	end
	
	-- classifications
	if f.level.enabled then
		if f.boss:IsVisible() then
			f.level:SetText('??b')
			f.level:SetTextColor(1, 0, 0)
			f.level:Show()
		elseif f.state:IsVisible() then
			if f.state:GetTexture() == "Interface\\Tooltips\\EliteNameplateIcon"
			then
				f.level:SetText(f.level:GetText()..'+')
			else
				f.level:SetText(f.level:GetText()..'r')
			end
		end
	else
		f.level:Hide()
	end
	
	if f.state:IsVisible() then
		-- hide the elite/rare dragon
		f.state:Hide()
	end

	if not loadedNames[f.name.text] and
	   not f.guid
	then
		-- store this frame's name
		loadedNames[f.name.text] = f
	end
	
	---------------------------------------------- Trivial sizing/positioning --
	if f.firstChild:GetScale() < 1 then
		if not f.trivial then
			-- initialise trivial unit sizes
			if uiscale then
				f:SetSize(self:GetWidth()/uiscale, self:GetHeight()/uiscale)
			end
			
			local w,h = f:GetSize()
			local x,y = 
				floor((w / 2) - (kn.sizes.twidth / 2)),
				floor((h / 2) - (kn.sizes.theight / 2))
		
			f.health:ClearAllPoints()
			f.bg:ClearAllPoints()
			f.bg.fill:ClearAllPoints()
			f.name:ClearAllPoints()
			
			f.level:Hide()
			f.health.p:Hide()
			
			if f.health.mo then
				f.health.mo:Hide()
			end
			
			--SetFontSize(f.name, 'small')
			f.name:SetJustifyH('CENTER')
			
			f.name:SetPoint('BOTTOM', f.health, 'TOP', 0, -3)
						
			f.bg.fill:SetSize(kn.sizes.twidth, kn.sizes.theight)
			f.health:SetSize(kn.sizes.twidth-2, kn.sizes.theight-2)
			
			f.health:SetPoint('BOTTOMLEFT', x+1, y+1)
			
			f.bg.fill:SetPoint('BOTTOMLEFT', x, y)
			
			f.bg:SetPoint('BOTTOMLEFT', x-(kn.sizes.bgOffset-2), y-(kn.sizes.bgOffset-2))
			f.bg:SetPoint('TOPRIGHT', f, 'BOTTOMLEFT', x+kn.sizes.twidth+(kn.sizes.bgOffset-2), y+kn.sizes.theight+(kn.sizes.bgOffset-2))
		
			f.trivial = true
		else
			-- (performed each time a trivial frame is shown)
			f.level:Hide()
		end
	elseif f.trivial then
		-- return to normal sizes
		if uiscale then
			f:SetSize(self:GetWidth()/uiscale, self:GetHeight()/uiscale)
		end
		
		local w,h = f:GetSize()
		local x,y =
			floor((w / 2) - (kn.sizes.width / 2)), 
			floor((h / 2) - (kn.sizes.height / 2))
		
		f.health:ClearAllPoints()
		f.bg:ClearAllPoints()
		f.bg.fill:ClearAllPoints()
		f.name:ClearAllPoints()
			
		f.health.p:Show()	
		
		if f.health.mo then
			f.health.mo:Show()
		end
		
		--SetFontSize(f.name, 'name')
		f.name:SetJustifyH('LEFT')
		f.name:SetPoint('RIGHT', f.health.p, 'LEFT')
		
		if f.level.enabled then
			f.name:SetPoint('LEFT', f.level, 'RIGHT', -2, 0)
			f.level:Show()
		else
			f.name:SetPoint('BOTTOMLEFT', f.health, 'TOPLEFT', 2, uiscale and -(2/uiscale) or -2)
		end
		
		f.bg.fill:SetSize(kn.sizes.width, kn.sizes.height)		
		f.health:SetSize(kn.sizes.width - 2, kn.sizes.height - 2)
		
		f.health:SetPoint('BOTTOMLEFT', x+1, y+1)
		
		f.bg.fill:SetPoint('BOTTOMLEFT', x, y)
		
		f.bg:SetPoint('BOTTOMLEFT', x-kn.sizes.bgOffset, y-kn.sizes.bgOffset)
		f.bg:SetPoint('TOPRIGHT', f, 'BOTTOMLEFT', x+(kn.sizes.width)+kn.sizes.bgOffset, y+(kn.sizes.height)+kn.sizes.bgOffset)
		
		f.trivial = nil
	end
	
	-- force health update
	OnHealthValueChanged(f.oldHealth, f.oldHealth:GetValue())
	f:SetHealthColour()
	f:SetGlowColour()

	if f.fixaa then
		f.DoShow = true
	else
		f:Show()
	end

	kn:SendMessage('KuiNameplates_PostShow', f)
end

local function OnFrameHide(self)
	local f = self.kui
	f:Hide()

	f:SetFrameLevel(1)
	
	if f.targetGlow then
		f.targetGlow:Hide()
	end

	if f.guid and loadedGUIDs[f.guid] == f then
		-- remove guid from the store
		loadedGUIDs[f.guid] = nil
	end

	f.guid = nil

	if loadedNames[f.name.text] == f then
		-- remove name from store
		-- if there are name duplicates, this will be recreated in an onupdate
		loadedNames[f.name.text] = nil
	end

	f.lastAlpha	= 0
	f.fadingTo	= nil
	f.hasThreat	= nil
	f.target	= nil

	-- unset stored health bar colours
	f.health.r, f.health.g, f.health.b, f.health.reset
		= nil, nil, nil, nil
	
	kn:SendMessage('KuiNameplates_PostHide', f)	
end

local function OnFrameEnter(self)
	self:StoreGUID(UnitGUID('mouseover'))

	if self.highlight then
		self.highlight:Show()
	end

	if profile.hp.mouseover then
		self.health.p:Show()
		if self.health.mo then self.health.mo:Show() end
	end
end

local function OnFrameLeave(self)
	if self.highlight then
		self.highlight:Hide()
	end

	if not self.target and profile.hp.mouseover then
		self.health.p:Hide()
		if self.health.mo then self.health.mo:Hide() end
	end
end

-- stuff that needs to be updated every frame
local function OnFrameUpdate(self, e)
	local f = self.kui

	f.elapsed	= f.elapsed - e
	f.critElap	= f.critElap - e

	if f.fixaa then
		------------------------------------------------------------ Position --
		local scale = f.firstChild:GetScale()
		local x, y = select(4, f.firstChild:GetPoint())
		x = (x / uiscale) * scale
		y = (y / uiscale) * scale
	
		f:SetPoint('BOTTOMLEFT', WorldFrame, 'BOTTOMLEFT',
			floor(x - (f:GetWidth() / 2)),
			floor(y))
		
		-- show the frame after it's been moved so it doesn't flash
		-- .DoShow is set OnFrameShow
		if f.DoShow then
			f:Show()
			f.DoShow = nil
		end
	end
	
	f.defaultAlpha = self:GetAlpha()

	------------------------------------------------------------------- Alpha --
	if (f.defaultAlpha == 1 and
	    targetExists)          or
	   (profile.fade.fademouse and
	    f.highlighted)
	then
		f.currentAlpha = 1
	elseif	targetExists or profile.fade.fadeall then
		f.currentAlpha = profile.fade.fadedalpha or .3
	else
		f.currentAlpha = 1
	end
	------------------------------------------------------------------ Fading --
	if profile.fade.smooth then
		-- track changes in the alpha level and intercept them
		if f.currentAlpha ~= f.lastAlpha then
			if not f.fadingTo or f.fadingTo ~= f.currentAlpha then
				if kui.frameIsFading(f) then
					kui.frameFadeRemoveFrame(f)
				end

				-- fade to the new value
				f.fadingTo 	  = f.currentAlpha
				local alphaChange = (f.fadingTo - (f.lastAlpha or 0))

				kui.frameFade(f, {
					mode		= alphaChange < 0 and 'OUT' or 'IN',
					timeToFade	= abs(alphaChange) * (profile.fade.fadespeed or .5),
					startAlpha	= f.lastAlpha or 0,
					endAlpha	= f.fadingTo,
					finishedFunc = function()
						f.fadingTo = nil
					end,
				})
			end

			f.lastAlpha = f.currentAlpha
		end
	else
		f:SetAlpha(f.currentAlpha)
	end

	-- call delayed updates
	if f.elapsed <= 0 then
		f.elapsed = slowUpdateTime
		f:UpdateFrame()
	end

	if f.critElap <= 0 then
		f.critElap = critUpdateTime
		f:UpdateFrameCritical()
	end
end

-- stuff that can be updated less often
local function UpdateFrame(self)
	if not loadedNames[self.name.text] and
	   not self.guid
	then
		-- ensure a frame is still stored for this name, as name conflicts cause
		-- it to be erased when another might still exist
		-- also ensure that if this frame is targeted, this is the stored frame
		-- for its name
		loadedNames[self.name.text] = self
	end

	-- Health bar colour
	self:SetHealthColour()
	
	-- Name text colour
	self:SetNameColour()
end

-- stuff that needs to be updated often
local function UpdateFrameCritical(self)
	------------------------------------------------------------------ Threat --
	if self.glow:IsVisible() then
		self.glow.wasVisible = true

		-- set glow to the current default ui's colour
		self.glow.r, self.glow.g, self.glow.b = self.glow:GetVertexColor()
		self:SetGlowColour(self.glow.r, self.glow.g, self.glow.b)

		if not self.friend and profile.tank.enabled then
			-- in tank mode; is the default glow red (are we tanking)?
			self.hasThreat = (self.glow.g + self.glow.b) < .1

			if self.hasThreat then
				-- tanking; recolour bar & glow
				local r, g, b, a = unpack(profile.tank.glowcolour)
				self:SetGlowColour(r, g, b, a)
				self:SetHealthColour()
			end
		end
	elseif self.glow.wasVisible then
		self.glow.wasVisible = nil

		-- restore shadow glow colour
		self:SetGlowColour()

		if self.hasThreat then
			-- lost threat
			self.hasThreat = nil
			self:SetHealthColour()
		end
	end
	------------------------------------------------------------ Target stuff --
	if targetExists and
	   self.defaultAlpha == 1 and
	   self.name.text == UnitName('target')
	then
		if not self.target then
			-- this frame just became targeted
			self.target = true
			self:StoreGUID(UnitGUID('target'))

			-- move this frame above others
			-- default UI uses a level of 10 by default & 20 on the target
			self:SetFrameLevel(10)

			if profile.hp.mouseover then
				self.health.p:Show()
				if self.health.mo then self.health.mo:Show() end
			end
		
			if self.targetGlow then
				self.targetGlow:Show()
			end

			kn:SendMessage('KuiNameplates_PostTarget', self)
		end
	elseif self.target then
		self.target = nil

		self:SetFrameLevel(1)
		
		if self.targetGlow then
			self.targetGlow:Hide()
		end
		
		if not self.highlighted and profile.hp.mouseover then
			self.health.p:Hide()
			if self.health.mo then self.health.mo:Hide() end
		end
	end

	--------------------------------------------------------------- Mouseover --
	if self.oldHighlight:IsShown() then
		if not self.highlighted then
			self.highlighted = true
			OnFrameEnter(self)
		end
	elseif self.highlighted then
		self.highlighted = false
		OnFrameLeave(self)
	end
	
	-- [debug]
	if _G['KuiNameplatesDebug'] then
		if self.guid then
			self.guidtext:SetText(self.guid)

			if loadedGUIDs[self.guid] ~= self then
				self.guidtext:SetTextColor(1,0,0)
			else
				self.guidtext:SetTextColor(1,1,1)
			end
		else
			self.guidtext:SetText(nil)
		end

		if self.name.text and loadedNames[self.name.text] == self then
			self.nametext:SetText('Has name')
		else
			self.nametext:SetText(nil)
		end
		
		if self.friend then
			self.isfriend:SetText('friendly')
		else
			self.isfriend:SetText('not friendly')
		end
	end
end

--------------------------------------------------------------- KNP functions --
function kn:GetNameplate(guid, name)
	local gf, nf = loadedGUIDs[guid], loadedNames[name]

	if gf then
		return gf
	elseif nf then
		return nf
	else
		return nil
	end
end

function kn:IsNameplate(frame)
	if frame:GetName() and strfind(frame:GetName(), '^NamePlate%d') then
		local nameTextChild = select(2, frame:GetChildren())
		if nameTextChild then
			local nameTextRegion = nameTextChild:GetRegions()
			return (nameTextRegion and nameTextRegion:GetObjectType() == 'FontString')
		end
	end
end

function kn:InitFrame(frame)
	-- container for kui objects!
	frame.kui = CreateFrame('Frame', nil, WorldFrame) 
	local f = frame.kui

	f.fontObjects = {}
	
	-- fetch default ui's objects
	local overlayChild, nameTextChild = frame:GetChildren()
	local healthBar, castBar = overlayChild:GetChildren()
	
	local _, castbarOverlay, shieldedRegion, spellIconRegion
		= castBar:GetRegions()

	local nameTextRegion = nameTextChild:GetRegions()
    local glowRegion, overlayRegion, highlightRegion, levelTextRegion,
          bossIconRegion, raidIconRegion, stateIconRegion
		= overlayChild:GetRegions()

	overlayRegion:SetTexture(nil)
	highlightRegion:SetTexture(nil)
	bossIconRegion:SetTexture(nil)
	shieldedRegion:SetTexture(nil)
	castbarOverlay:SetTexture(nil)
	glowRegion:SetTexture(nil)

	-- make default healthbar & castbar transparent
	healthBar:SetStatusBarTexture(kui.m.t.empty)
	castBar:SetStatusBarTexture(kui.m.t.empty)

	f.firstChild = overlayChild

	--f.bg    = overlayRegion
	f.glow  = glowRegion
	f.boss  = bossIconRegion
	f.state = stateIconRegion
	f.level = levelTextRegion
	f.icon  = raidIconRegion
	f.spell = spellIconRegion
	f.oldHealth = healthBar

	f.oldName = nameTextRegion
	f.oldName:Hide()
	
	f.oldHighlight = highlightRegion

    --------------------------------------------------------- Frame functions --
    f.CreateFontString    = CreateFontString
    f.UpdateFrame         = UpdateFrame
    f.UpdateFrameCritical = UpdateFrameCritical
    f.SetHealthColour     = SetHealthColour
    f.SetNameColour       = SetNameColour
    f.SetGlowColour       = SetGlowColour
	f.StoreGUID           = StoreFrameGUID

    ------------------------------------------------------------------ Layout --
	local parent
	if profile.general.fixaa and uiscale then
		f:SetFrameStrata('BACKGROUND')
		f:SetSize(frame:GetWidth()/uiscale, frame:GetHeight()/uiscale)
		f:SetScale(uiscale)
		
		f:SetPoint('BOTTOMLEFT', UIParent)
		f:Hide()
		
		-- [debug]
		if _G['KuiNameplatesDebug'] then
			f:SetBackdrop({ bgFile = kui.m.t.solid })
			f:SetBackdropColor(0,0,0,.5)
		end
		
		f.fixaa = true
	else
		f:SetAllPoints(frame)
	end

	parent = f
	f.parent = parent
	
	-- using CENTER breaks pixel-perfectness with oddly sized frames
	-- .. so we have to align frames manually.
	local w,h = f:GetSize()
	f.x = floor((w / 2) - (kn.sizes.width / 2))
	f.y = floor((h / 2) - (kn.sizes.height / 2))

	self:CreateBackground(frame, f)
	self:CreateHealthBar(frame, f)

	-- overlay (text is parented to this) --------------------------------------
	f.overlay = CreateFrame('Frame', nil, f)
	f.overlay:SetAllPoints(f.health)
	f.overlay:SetFrameLevel(f.health:GetFrameLevel()+1)

	self:CreateHighlight(frame, f)
	self:CreateHealthText(frame, f)
	
	if profile.hp.showalt then
		self:CreateAltHealthText(frame, f)
	end

	if profile.text.level then
		self:CreateLevel(frame, f)
	else
		f.level:Hide()
	end

	self:CreateName(frame, f)
	
	-- target highlight --------------------------------------------------------
	if profile.general.targetglow then
		f.targetGlow = f.overlay:CreateTexture(nil, 'ARTWORK')
		f.targetGlow:SetTexture('Interface\\AddOns\\Kui_Nameplates\\media\\target-glow')
		f.targetGlow:SetTexCoord(0, .593, 0, .875)
		f.targetGlow:SetPoint('TOP', f.overlay, 'BOTTOM', 0, 1)
		f.targetGlow:SetVertexColor(unpack(profile.general.targetglowcolour))
		f.targetGlow:SetSize(kn.texSizes.targetGlowW, kn.texSizes.targetGlowH)
		f.targetGlow:Hide()
	end

	-- raid icon ---------------------------------------------------------------
	f.icon:SetParent(f.overlay)
	f.icon:SetSize(kn.texSizes.raidicon, kn.texSizes.raidicon)

	f.icon:ClearAllPoints()
	f.icon:SetPoint('LEFT', f.health, 'RIGHT', 5, 1)
    ----------------------------------------------------------------- Scripts --
	frame:HookScript('OnShow', OnFrameShow)
	frame:HookScript('OnHide', OnFrameHide)
	frame:HookScript('OnUpdate', OnFrameUpdate)

	f.oldHealth:HookScript('OnValueChanged', OnHealthValueChanged)

	-- [debug]
	if _G['KuiNameplatesDebug'] then
		frame:SetBackdrop({bgFile=kui.m.t.solid})
		frame:SetBackdropColor(1, 1, 1, .5)

		f.isfriend = f:CreateFontString(f.overlay)
		f.isfriend:SetPoint('BOTTOM', frame, 'TOP')
		
		f.guidtext = f:CreateFontString(f.overlay)
		f.guidtext:SetPoint('TOP', frame, 'BOTTOM')

		f.nametext = f:CreateFontString(f.overlay)
		f.nametext:SetPoint('TOP', f.guidtext, 'BOTTOM')
	end

	------------------------------------------------------------ Finishing up --
    f.elapsed  = slowUpdateTime
	f.critElap = critUpdateTime

	kn:SendMessage('KuiNameplates_PostCreate', f)
	
	if frame:IsShown() then
		-- force OnShow
		OnFrameShow(frame)
	else
		f:Hide()
	end
end

---------------------------------------------------------------------- Events --
function kn:PLAYER_TARGET_CHANGED()
	targetExists = UnitExists('target')
end

-- automatic toggling of enemy frames
function kn:PLAYER_REGEN_ENABLED()
	SetCVar('nameplateShowEnemies', 0)
end
function kn:PLAYER_REGEN_DISABLED()
	SetCVar('nameplateShowEnemies', 1)
end

------------------------------------------------------------- Script handlers --
kn.numFrames = 0

do
	local WorldFrame = WorldFrame

	function kn:OnUpdate()
		local frames = select('#', WorldFrame:GetChildren())

		if frames ~= self.numFrames then
			local i, f

			for i = 1, frames do
				f = select(i, WorldFrame:GetChildren())
				if self:IsNameplate(f) and not f.kui then
					self:InitFrame(f)
					tinsert(self.frameList, f)
				end
			end

			self.numFrames = frames
		end
	end

	kn:ScheduleRepeatingTimer('OnUpdate', .1)
end

function kn:ToggleCombatEvents(io)
	if io then
		self:RegisterEvent('PLAYER_REGEN_ENABLED')
		self:RegisterEvent('PLAYER_REGEN_DISABLED')
	else
		self:UnregisterEvent('PLAYER_REGEN_ENABLED')
		self:UnregisterEvent('PLAYER_REGEN_DISABLED')
	end
end

-------------------------------------------------- Listen for profile changes --
function kn:ProfileChanged()
	profile = self.db.profile
end

---------------------------------------------------- Post db change functions --
-- n.b. this is absolutely terrible and horrible and i hate it
kn.configChangedFuncs.runOnce.fontscale = function(val)
	kn:ScaleFontSizes()
end
kn.configChangedFuncs.fontscale = function(frame, val)
	local _, fontObject
	for _, fontObject in pairs(frame.fontObjects) do
		if type(fontObject.size) == 'string' then
			SetFontSize(fontObject, kn.fontSizes[fontObject.size])
		end
	end
end

kn.configChangedFuncs.outline = function(frame, val)
	local _, fontObject
	for _, fontObject in pairs(frame.fontObjects) do
		kui.ModifyFontFlags(fontObject, val, 'OUTLINE')
	end
end

kn.configChangedFuncs.monochrome = function(frame, val)
	local _, fontObject
	for _, fontObject in pairs(frame.fontObjects) do
		kui.ModifyFontFlags(fontObject, val, 'MONOCHROME')
	end
end

kn.configChangedFuncs.fontscale = function(frame, val)
	local _, fontObject
	for _, fontObject in pairs(frame.fontObjects) do
		if type(fontObject.size) == 'string' then
			SetFontSize(fontObject, kn.fontSizes[fontObject.size])
		end
	end
end
kn.configChangedFuncs.onesize = kn.configChangedFuncs.fontscale

kn.configChangedFuncs.Health = function(frame)
	if frame:IsShown() then
		-- update health display
		OnHealthValueChanged(frame.oldHealth, frame.oldHealth:GetValue())
	end
end
kn.configChangedFuncs.friendly = kn.configChangedFuncs.Health
kn.configChangedFuncs.hostile = kn.configChangedFuncs.Health

kn.configChangedFuncs.runOnce.font = function(val)
	kn.font = LSM:Fetch(LSM.MediaType.FONT, val)
end
kn.configChangedFuncs.font = function(frame, val)
	local _, fontObject
	for _, fontObject in pairs(frame.fontObjects) do
		local _, size, flags = fontObject:GetFont()
		fontObject:SetFont(kn.font, size, flags)
	end
end

-------------------------------------------------------------------- Finalise --
-- scale a frame/font size to keep it relatively the same with any uiscale
function ScaleFrameSize(key)
	local size = kn.defaultSizes.frame[key]
	kn.sizes[key] = (uiscale and floor(size/uiscale) or size)
end

function ScaleTextureSize(key)
	-- texture sizes don't need to be rounded
	local size = kn.defaultSizes.tex[key]
	kn.texSizes[key] = (uiscale and size/uiscale or size)
end

function ScaleFontSize(key)
	local size = kn.defaultSizes.font[key]
	kn.fontSizes[key] = size * profile.fonts.options.fontscale
		
	if uiscale then
		kn.fontSizes[key] = kn.fontSizes[key] / uiscale
	end
end

-- [initialise and] scale sizes
function kn:ScaleFrameSizes()
	local k,_
	for k,_ in pairs(kn.defaultSizes.frame) do
		ScaleFrameSize(k)
	end
end
function kn:ScaleTextureSizes()
	local k,_
	for k,_ in pairs(kn.defaultSizes.tex) do
		ScaleTextureSize(k)
	end
end
function kn:ScaleFontSizes()
	local k,_
	for k,_ in pairs(kn.defaultSizes.font) do
		ScaleFontSize(k)
	end
end

-- modules must use this to add correctly scaled sizes
-- font sizes can then be called as a key in kn.CreateFontString
function kn:RegisterSize(type, key, size)
	if not kn.defaultSizes[type] then return end

	kn.defaultSizes[type][key] = size

	if type == 'frame' then
		ScaleFrameSize(key)
	elseif type == 'tex' then
		ScaleTextureSize(key)
	else
		ScaleFontSize(key)
	end
end

function kn:OnEnable()
	profile = self.db.profile
    
	if profile.general.fixaa then
		uiscale = UIParent:GetEffectiveScale()
	end
		
	self:ScaleFrameSizes()
	self:ScaleTextureSizes()
	self:ScaleFontSizes()
	
	-- fetch font path from lsm
	self.font = LSM:Fetch(LSM.MediaType.FONT, profile.fonts.options.font)
    
	-------------------------------------- Health bar smooth update functions --
	-- (spoon-fed by oUF_Smooth)
	if profile.hp.smooth then
		local f, smoothing, GetFramerate, min, max, abs
			= CreateFrame('Frame'), {}, GetFramerate, math.min, math.max, math.abs

		function kn.SetValueSmooth(self, value)
			local _, maxv = self:GetMinMaxValues()

			if value == self:GetValue() or (self.prevMax and self.prevMax ~= maxv) then
				-- finished smoothing/max health updated
				smoothing[self] = nil
				self:OrigSetValue(value)
			else
				smoothing[self] = value
			end

			self.prevMax = maxv
		end

		f:SetScript('OnUpdate', function()
			local limit = 30/GetFramerate()
			
			for bar, value in pairs(smoothing) do
				local cur = bar:GetValue()
				local new = cur + min((value-cur)/3, max(value-cur, limit))

				if new ~= new then
					new = value
				end

				bar:OrigSetValue(new)

				if cur == value or abs(new - value) < 2 then
					bar:OrigSetValue(value)
					smoothing[bar] = nil
				end
			end
		end)
	end

	self:RegisterEvent('PLAYER_TARGET_CHANGED')
	self:ToggleCombatEvents(profile.general.combat)
end
